#ifndef lint
static const char rcsid[] =
	"$Id: password.c,v 1.1.1.1 2001/01/12 06:06:44 andrew Exp $";
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "config.h"
#include "password.h"
#include "poppwd.h"

int authenticate(const char *const username, const char *const pass) {
	struct passwd *pw;

	if ((pw = getpwnam(username)) == NULL) {
		/* no such user */
		return(0);
	}

	/* disallow attempts to change system passwords */
	if (pw->pw_uid < (uid_t)MIN_UID) {
		/* system user */
		syslog(LOG_NOTICE, "attempt to login as system user %s", username);
		return(0);
	}

	if (strcmp(crypt(pass, pw->pw_passwd), pw->pw_passwd) == 0) {
		return(1);
	} else {
		return(0);
	}
}

int change_pass(const char *const username, const char *const pass) {
	pid_t pid;
	int fds[2],
		status;
	char fd_name[FD_MAXLENGTH + 1];

	/* grab a fd to talk to pw over */
	if (pipe(fds) != 0) {
		basic_err(EX_OSERR, MSG_ERR_MISC, "pipe failed: %m");
	}
	/* convert fd[0] to ascii */
	if (snprintf(fd_name, (size_t)(FD_MAXLENGTH + 1), "%d", fds[0]) >= (FD_MAXLENGTH)) {
		basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "snprintf failed: %m");
	}

	/* fork so we can exec safely */
	switch (pid = fork()) {
	case -1:
		/* error */
		basic_err(EX_OSERR, MSG_ERR_MISC, "fork failed: %m");
		break;
	case 0:
		/* child */
		if (execl(PW, "pw", "usermod", username, "-h", fd_name, NULL) == -1) {
			syslog(LOG_ERR, "execl of %s of behalf of process %ld failed: %m", PW, (long)getppid());
			/* parent will tell client */
			exit(EX_CONFIG);
		}
		break;
	default:
		/* parent */
		if (write(fds[1], pass, strlen(pass)) < 1) {
			basic_err(EX_IOERR, MSG_ERR_IO, "write to pw failed: %m");
		}
		if (wait(&status) == -1) {
			basic_err(EX_OSERR, MSG_ERR_MISC, "wait failed: %m");
		} else {
			if (WIFEXITED(status)) {
				if (WEXITSTATUS(status) == EX_OK) {
					/* worked ok */
					return(1);
				} else {
					basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "pw exited with code %d", WEXITSTATUS(status));
				}
			} else if (WIFSIGNALED(status)) {
				basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "pw exited on signal %d", WTERMSIG(status));
			} else {
				basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "pw exited abnormally");
			}
		}
		break;
	}

	return(0);
}
