#ifndef lint
static const char rcsid[] =
	"$Id: protocol.c,v 1.1.1.1 2001/01/12 06:06:44 andrew Exp $";
#endif

#include <string.h>
#include <sysexits.h>
#include <syslog.h>

#include "config.h"
#include "password.h"
#include "poppwd.h"
#include "protocol.h"

#include "protocol_private.h"

void do_protocol(void) {
	char *username,
		 *old_pass,
		 *new_pass,
		 *arg;
	enum { WANT_USER,
		   WANT_OLDPASS,
		   WANT_NEWPASS,
		   DONE } state;
	enum command cmd;

	/* set the initial state */
	state = WANT_USER;

	/* print welcome banner */
	basic_to_client(MSG_WELCOME);

	for (;;) {
		(void)lex(&cmd, &arg);
		/* actions not dependant on state */
		switch (cmd) {
		case CMD_UNKNOWN:
			basic_to_client(MSG_ERR_PROTO);
			continue;
			break;
		case CMD_QUIT:
			basic_to_client(MSG_GOODBYE);
			return;
			break;
		case CMD_USER:
		case CMD_PASS:
		case CMD_NEWPASS:
			/* ignore these as they depend on state */
			break;
		}

		switch (state) {
		case WANT_USER:
			if ((cmd == CMD_USER) && ((username = arg) != NULL)) {
				basic_to_client(MSG_OLDPASS);
				state = WANT_OLDPASS;
			} else {
				basic_to_client(MSG_ERR_PROTO);
			}
			break;
		case WANT_OLDPASS:
			if ((cmd == CMD_PASS) && ((old_pass = arg) != NULL)) {
				if (authenticate(username, old_pass)) {
					syslog(LOG_INFO, "login by %s", username);
					basic_to_client(MSG_NEWPASS);
					state = WANT_NEWPASS;
				} else {
					syslog(LOG_INFO, "failed attempt to authenticate %s", username);
					basic_to_client(MSG_ERR_AUTH);
					state = DONE;
				}
			} else {
				basic_to_client(MSG_ERR_PROTO);
			}
			break;
		case WANT_NEWPASS:
			if ((cmd == CMD_NEWPASS) && ((new_pass = arg) != NULL)) {
				if (change_pass(username, new_pass)) {
					/* worked */
					syslog(LOG_NOTICE, "password changed for %s", username);
					basic_to_client(MSG_SUCCESS);
					state = DONE;
				} else {
					/* failure */
					basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "change_pass failed: %m");
				}
			} else {
				basic_to_client(MSG_ERR_PROTO);
			}
			break;
		case DONE:
		default:
			basic_to_client(MSG_ERR_PROTO);
			break;
		}
	}
}

int lex(enum command *cmd, char **arg) {
	char buffer[MAX_LINE + 1]; /* a place to store incoming data */
	char *argptr;
	unsigned int i;
				
	/* read line */
	from_client(TIME_IN, MAX_LINE + 1, buffer);

	/* try to match a command */
	*cmd = CMD_UNKNOWN;
	for (i = 0; i < (sizeof(commands) / sizeof(const char *const)); ++i) {
		if (strncasecmp(buffer, commands[i], strlen(commands[i])) == 0) {
			/* we match the first bit */
			argptr = buffer + strlen(commands[i]);
			if (strchr("\r\n\t ", *argptr) != NULL) {
				/* followed by whitespace so we have a match */
				*cmd = i;
				break;
			}
		}
	}

	if (*cmd != CMD_UNKNOWN) {
		/* strip leading whitespace from argument */
		for (; strchr("\t ", *argptr) != NULL; ++argptr) {
			/* This space intentionally left blank */
		}
		/* strip off \r\n allowing for the \r to be missing */
		*strpbrk(buffer, "\r\n") = '\0'; /* we are guarenteed a \n by
										  * from_client */
		if (*argptr != '\0') {
			if ((*arg = strdup(argptr)) == NULL) {
				basic_err(EX_UNAVAILABLE, MSG_ERR_MISC, "strdup failed: %m");
			}
		} else {
			*arg = NULL;
		}
	}

	return(1);
}
